/* This class is to compare two oracle-info files to see if they match
 * and, if not, to point to a location at which they mismatch. 
 * 
 * Note that this is not a comparison in the sense of
 * java.lang.Comparable.
 *
 * Important: Only mismatch information is ever printed to System.out
 * by this class. All other information (e.g., errors) must go to
 * System.err.
 */

package oracle; 

import jargs.gnu.CmdLineParser; 
import java.io.File; 
import java.io.IOException; 
import org.jdom.JDOMException; 
import java.util.*; 


public class Comparer {

    /* Specifies which steps should be compared-- all steps
     * (PROC_ALL_STEPS) or only the last step (PROC_LAST_STEP).
     */
    private int oracleProcedure; 

    /* @see oracleProcedure
     */
    public static final int PROC_ALL_STEPS = 0, PROC_LAST_STEP = 1; 

    /* Specifies which components should be compared-- all components
     * in all windows (INFO_ALL_COMPONENTS), the component acted on in
     * the step (INFO_EVENT_COMPONENT), or the components in the
     * active window (INFO_ACTIVE_WINDOW).
     */
    private int oracleInfo; 

    /* @see oracleInfo
     */
    public static final int INFO_ALL_COMPONENTS = 0, INFO_EVENT_COMPONENT = 1, INFO_ACTIVE_WINDOW = 2; 

    /* If true, a mismatch is returned for each step that has a
     * mismatch. If false, only a mismatch for the first mismatching
     * step is returned.
     */
    private boolean eachStepMismatch; 

    /* Parts to ignore of each oracle-info file that gets compared.
     */
    private Ignore ignore; 

    public Comparer(File ignoreFile, int oracleProcedure, int oracleInfo, boolean eachStepMismatch) throws IllegalArgumentException, IOException, JDOMException {
	if (oracleProcedure < PROC_ALL_STEPS || oracleProcedure > PROC_LAST_STEP) {
	    throw new IllegalArgumentException("Invalid oracle procedure");
	}
	if (oracleInfo < INFO_ALL_COMPONENTS || oracleInfo > INFO_ACTIVE_WINDOW) {
	    throw new IllegalArgumentException("Invalid oracle info"); 
	}
	this.oracleProcedure = oracleProcedure; 
	this.oracleInfo = oracleInfo; 
	this.eachStepMismatch = eachStepMismatch; 
	ignore = new Ignore(ignoreFile); 
    }

    /* If eachStepMismatch is true, returns a vector with the same
     * number of elements as there are complete steps in oracleFile1
     * or oracleFile2 (whichever is greater). Each element is either a
     * Mismatch for that step or null if there is no mismatch (<i>for
     * that step</i>).
     *
     * If eachStepMismatch is false, returns a vector of one element, either
     * the first Mismatch found or null if there is no mismatch for
     * any step.
     */
    public Vector<Mismatch> compare(File oracleFile1, File oracleFile2) throws IOException, JDOMException, UnimplementedException, OracleReader.XMLStructureException {
	if (oracleProcedure != PROC_ALL_STEPS) {
	    throw new UnimplementedException("Oracle procedure " + oracleProcedure + " has not been implemented"); 
	}
	OracleReader input1 = new OracleReader(oracleFile1, ignore); 
	OracleReader input2 = new OracleReader(oracleFile2, ignore); 
	Step step1, step2; 
	Vector<Mismatch> mismatches = new Vector<Mismatch>();
	do {
	    step1 = input1.nextStep(); 
	    step2 = input2.nextStep(); 
	    if (step1 != null && step2 != null) { 
		Mismatch m = compare(step1, step2); 
		if (!eachStepMismatch) {
		    if (m != null) {
			mismatches.addElement(m); 
			return mismatches; 
		    }
		} else {
		    mismatches.addElement(m); 
		}
	    } else {
		if (step1 == null && step2 != null) {
		    Mismatch m = new Mismatch(Mismatch.TYPE_A, Mismatch.LOC_STEP, step2.getNumber(), step2.getEvent(), null, null, null); 
		    mismatches.addElement(m); 
		    if (!eachStepMismatch) {
			return mismatches; 
		    }
		} else if (step1 != null && step2 == null) {
		    Mismatch m = new Mismatch(Mismatch.TYPE_D, Mismatch.LOC_STEP, step1.getNumber(), step1.getEvent(), null, null, null); 
		    mismatches.addElement(m); 
		    if (!eachStepMismatch) {
			return mismatches; 
		    }
		}
	    }
	} while (step1 != null || step2 != null); 
	return mismatches; 
    }

    /* Returns null if there is no mismatch. 
     */
    public Mismatch compare(Step step1, Step step2) throws JDOMException, UnimplementedException {
	if (oracleInfo != INFO_EVENT_COMPONENT && oracleInfo != INFO_ALL_COMPONENTS) {
	    throw new UnimplementedException("Oracle info level " + oracleInfo + " has not been implemented"); 
	}
	Mismatch m; 
	m = compare(step1.getEvent(), step2.getEvent()); 
	if (m != null) {
	    m.setStep(step1.getNumber()); 
	    return m; 
	}
	if (oracleInfo == INFO_EVENT_COMPONENT) { 
	    m = compare(step1.getCompInfo(), step2.getCompInfo()); 
	    if (m != null) {
		m.setStep(step1.getNumber()); 
		m.setEvent(step1.getEvent()); 
		return m; 
	    }
	} else if (oracleInfo == INFO_ALL_COMPONENTS) {
	    Vector<Window> unpaired = new Vector<Window>(step2.getAllInfo()); 
	    for (Window w1 : step1.getAllInfo()) {
		Vector<String> pairingProperties = new Vector(w1.keySet()); 
		Collections.sort(pairingProperties); 
		pairingProperties.remove("Title"); 
		pairingProperties.remove("Class"); 
		pairingProperties.add(0, "Title"); 
		pairingProperties.add(1, "Class"); 
		Vector<Window> possPairings = new Vector<Window>(unpaired); 
		for (String prop : pairingProperties) {
		    if (possPairings.size() <= 1) {
			break; 
		    }
		    for (Iterator<Window> i = possPairings.iterator(); i.hasNext(); ) {
			Window w2 = i.next(); 
			if (! w1.get(prop).equals(w2.get(prop))) {
			    i.remove(); 
			}
		    }
		}
		if (possPairings.size() == 0 || ! possPairings.get(0).get("Title").equals(w1.get("Title"))) {
		    return new Mismatch(Mismatch.TYPE_D, Mismatch.LOC_WIN, step1.getNumber(), step1.getEvent(), w1, null, null); 
		}
		Window w2 = possPairings.get(0); 
		unpaired.remove(w2); 
		m = compare(w1, w2); 
		if (m != null) {
		    m.setStep(step1.getNumber()); 
		    m.setEvent(step1.getEvent()); 
		    return m;
		}

	    }
	    if (unpaired.size() > 0) {
		return new Mismatch(Mismatch.TYPE_A, Mismatch.LOC_WIN, step1.getNumber(), step1.getEvent(), unpaired.get(0), null, null); 
	    }
	}

	return null; 
    }

    /* Returns null if there is no mismatch. Otherwise returns a
     * Mismatch for which step needs to be filled in.
     */
    public Mismatch compare(Event event1, Event event2) {
	Vector<String> properties = new Vector<String>(event1.keySet()); 
	Collections.sort(properties); 
	for (String prop : properties) {
	    if (! event1.get(prop).equals(event2.get(prop))) {
		return new Mismatch(Mismatch.TYPE_C, Mismatch.LOC_STEP, -1, event1, null, null, event2); 
	    }
	}

	return null; 
    }


    /* Returns null if there is no mismatch. Otherwise returns a
     * Mismatch for which step, event, and (if applicable) win need to
     * be filled in.
     */
    public Mismatch compare(Component comp1, Component comp2) {
	Vector<String> properties = new Vector<String>(comp1.keySet()); 
	Collections.sort(properties); 
	for (String prop : properties) {
	    if (! comp1.get(prop).equals(comp2.get(prop))) {
		return new Mismatch(Mismatch.TYPE_C, Mismatch.LOC_COMP, -1, null, null, comp1, comp2); 
	    }
	}

	return null; 
    }

    /* Returns null if there is no mismatch. Otherwise returns a
     * Mismatch for which step and event need to be filled in.
     */
    public Mismatch compare(Window win1, Window win2) {
	Vector<String> properties = new Vector<String>(win1.keySet()); 
	Collections.sort(properties); 
	for (String prop : properties) {
	    if (! win1.get(prop).equals(win2.get(prop))) {
		return new Mismatch(Mismatch.TYPE_C, Mismatch.LOC_WIN, -1, null, win1, null, win2); 
	    }
	}
	Vector<Component> unpaired = new Vector<Component>(win2.getComponents()); 
	for (Component c1 : win1.getComponents()) {
	    Vector<String> pairingProperties = new Vector(c1.keySet()); 
	    Collections.sort(pairingProperties); 
	    pairingProperties.remove("Title"); 
	    pairingProperties.remove("Class"); 
	    pairingProperties.add(0, "Title"); 
	    pairingProperties.add(1, "Class"); 
	    Vector<Component> possPairings = new Vector<Component>(unpaired); 
	    for (String prop : pairingProperties) {
		if (possPairings.size() <= 1) {
		    break; 
		}
		for (Iterator<Component> i = possPairings.iterator(); i.hasNext(); ) {
		    Component c2 = i.next(); 
		    if (! c1.get(prop).equals(c2.get(prop))) {
			i.remove(); 
		    }
		}
	    }
	    if (possPairings.size() == 0 || ! possPairings.get(0).get("Title").equals(c1.get("Title"))) {
		return new Mismatch(Mismatch.TYPE_D, Mismatch.LOC_COMP, -1, null, win1, c1, null); 
	    }
	    Component c2 = possPairings.get(0); 
	    unpaired.remove(c2); 
	    Mismatch m = compare(c1, c2); 
	    if (m != null) {
		m.setWindow(win1); 
		return m; 
	    }
	}
	if (unpaired.size() > 0) {
	    return new Mismatch(Mismatch.TYPE_A, Mismatch.LOC_COMP, -1, null, win1, unpaired.get(0), null); 
	}
	
	return null; 
    }

    public static void main(String[] args) {
	String usage = 
	    "Usage: java oracle.Comparer [options] <oracle info file 1> <oracle info file 2>\n" + 
	    "Options include: \n" + 
	    "\t-i|--ignore <ignore file>\n" + 
	    "\t-p|--procedure <" + PROC_ALL_STEPS + " = check all steps (default), " + PROC_LAST_STEP + " = check last step>\n" + 
	    "\t-n|--info <" + INFO_ALL_COMPONENTS + " = check all components (default), " + INFO_EVENT_COMPONENT + " = check component acted on in step, " + INFO_ACTIVE_WINDOW + " = check active window>\n" + 
	    "\t-e|--each (return a mismatch for each mismatching step, instead of just the first mistmatching step)\n" + 
	    "If the compared files match, there is no output on standard out."; 

	CmdLineParser parser = new CmdLineParser(); 
	CmdLineParser.Option oIgnoreFileName = parser.addStringOption('i', "ignore"); 
	CmdLineParser.Option oOracleProcedure = parser.addIntegerOption('p', "procedure"); 
	CmdLineParser.Option oOracleInfo = parser.addIntegerOption('n', "info"); 
	CmdLineParser.Option oEachStepMismatch = parser.addBooleanOption('e', "each"); 
	try {
	    parser.parse(args); 
	    String ignoreFileName = (String) parser.getOptionValue(oIgnoreFileName); 
	    Integer oracleProcedureInt = (Integer) parser.getOptionValue(oOracleProcedure); 
	    Integer oracleInfoInt = (Integer) parser.getOptionValue(oOracleInfo); 
	    Boolean eachStepMismatchBool = (Boolean) parser.getOptionValue(oEachStepMismatch); 
	    File ignoreFile = (ignoreFileName == null) ? null : new File(ignoreFileName); 
	    int oracleProcedure = (oracleProcedureInt == null) ? PROC_ALL_STEPS : oracleProcedureInt.intValue(); 
	    int oracleInfo = (oracleInfoInt == null) ? INFO_ALL_COMPONENTS : oracleInfoInt.intValue(); 
	    boolean eachStepMismatch = (eachStepMismatchBool == null ) ? false : eachStepMismatchBool.booleanValue(); 
	    String[] oracleFileNames = parser.getRemainingArgs(); 
	    if (oracleFileNames.length != 2) {
		throw new IllegalArgumentException(); 
	    }
	    File oracleFile1 = new File(oracleFileNames[0]); 
	    File oracleFile2 = new File(oracleFileNames[1]); 

	    try {
		Comparer comparer = new Comparer(ignoreFile, oracleProcedure, oracleInfo, eachStepMismatch); 
		Vector<Mismatch> mismatches = comparer.compare(oracleFile1, oracleFile2); 
		for (Mismatch m : mismatches) {
		    if (m != null) {
			System.out.println(m.toString()); 
		    }
		}

	    } catch (IOException e) {
		e.printStackTrace(); 
	    } catch (JDOMException e) {
		e.printStackTrace(); 
	    } catch (OracleReader.XMLStructureException e) {
		e.printStackTrace(); 
	    } catch (UnimplementedException e) {
		System.err.println(e.getMessage()); 
	    }
	} catch (CmdLineParser.OptionException e) {
	    System.err.println(usage); 
	} catch (IllegalArgumentException e) {
	    System.err.println(usage); 
	}
    }

    protected class UnimplementedException extends Exception {

	public UnimplementedException(String s) {
	    super(s); 
	}

    }

}
