/* This class represents a mismatch (analogous to one element of
 * 'diff' output) between two oracle-info files.
 */

package oracle; 


public class Mismatch {

    /* The type of mismatch-- addition (TYPE_A), change (TYPE_C), or
     * deletion (TYPE_D).
     */
    private int type; 

    /* @see type
     */
    public static final int TYPE_A = 0, TYPE_C = 1, TYPE_D = 2; 

    /* String representations for possible values of location. 
     */
    public static final String[] NAMES_TYPE = {"add", "change", "delete"}; 

    /* The location or level at which the mismatch occurs-- step
     * (LOC_STEP), window (LOC_WIN), or component (LOC_COMP).
     */
    private int location; 

    /* @see location
     */
    public static final int LOC_STEP = 0, LOC_WIN = 1, LOC_COMP = 2; 

    /* String representations for possible values of location. 
     */
    public static final String[] NAMES_LOC = {"step", "window", "component"}; 

    /* The number of the test step at which the mismatch
     * occurred. 
     * 
     * @see Step.num
     */
    int step; 

    /* The event performed at the test step at which the mismatch
     * occurred. 
     */
    Event event; 

    /* The alternate event to event if the mismatch was a change in
     * event (otherwise null).
     */
    Event diffEvent; 

    /* The window at which the mismatch occurred if it was at the
     * window or somponent level (otherwise null).
     */
    Window win; 

    /* The alternate window to win if the mismatch was a change in
     * window (otherwise null).
     */
    Window diffWin; 

    /* The component at which the mismatch occurred if it was at the
     * component level (otherwise null).
     */
    Component comp; 

    /* The alternate component to comp if the mismatch was a change in
     * component (otherwise null).
     */
    Component diffComp; 

    public Mismatch(int type, int location, int step, Event event, Window win, Component comp, Object diffObj) throws IllegalArgumentException {
	if (type < TYPE_A || type > TYPE_D) {
	    throw new IllegalArgumentException("Invalid type"); 
	}
	if (location < LOC_STEP || location > LOC_COMP) {
	    throw new IllegalArgumentException("Invalid location"); 
	}
	this.type = type; 
	this.location = location; 
	this.step = step; 
	this.event = event; 
	this.win = win; 
	this.comp = comp; 
	if (type == TYPE_C) {
	    if (diffObj == null) {
		throw new IllegalArgumentException("diffObj must not be null for change type"); 
	    }
	    try {
		switch (location) { 
		    case LOC_STEP : diffEvent = (Event) diffObj; break; 
		    case LOC_WIN : diffWin = (Window) diffObj; break; 
		    case LOC_COMP : diffComp = (Component) diffObj; break; 
		}
	    } catch (ClassCastException e) {
		throw new IllegalArgumentException("Wrong type of diffObj for location " + NAMES_LOC[location]); 
	    }
	}
    }

    public String toString() {
	StringBuffer s = new StringBuffer(); 
	s.append("TYPE: " + NAMES_TYPE[type] + "\n" + 
		 "LOCATION: " + NAMES_LOC[location] + "\n" + 
		 "STEP: " + step + "\n" + 
		 "EVENT: " + event + "\n"); 
	if (location >= LOC_WIN) {
	    s.append("WINDOW: " + win + "\n"); 
	    if (location >= LOC_COMP) {
		s.append("COMPONENT: " + comp + "\n"); 
	    }
	}
	if (type == TYPE_C) {
	    switch (location) {
		case LOC_STEP: s.append("EVENT: " + diffEvent + "\n"); break; 
		case LOC_WIN: s.append("WINDOW: " + diffWin + "\n"); break; 
		case LOC_COMP: s.append("COMPONENT: " + diffComp + "\n"); break; 
	    } 
	}
	return s.toString(); 
    }

    public void setStep(int step) {
	this.step = step; 
    }

    public void setEvent(Event event) {
	this.event = event; 
    }

    public void setWindow(Window win) {
	this.win = win; 
    }

}
