/* A class for reading and parsing oracle-info files one step at a time. 
 */ 

package oracle; 

import java.io.*; 
import org.jdom.*; 
import org.jdom.input.*; 
import java.util.*; 

public class OracleReader {

    /* The file to read. 
     */
    private File oracleFile; 

    /* @see Comparer.ignore
     */
    Ignore ignore; 

    /* Status after reading the most recent step, or after
     * initialization if no steps have been read-- STATUS_PARTIAL if
     * the last step was cut off in the middle, STATUS_ABSENT if there
     * were no more steps left to read when the last step was
     * attempted, or STATUS_OK otherwise.
     */
    private int stepStatus; 

    /* @see stepStatus
     */
    public static final int STATUS_OK = 0, STATUS_PARTIAL = 1, STATUS_ABSENT = 2; 

    /* The number of the last step read. 
     * 
     * @see Step.number
     */
    private int stepNumber; 

    /* The reader of the oracle-info file. 
     */
    protected LineNumberReader reader; 

    /* The file reader around which reader is wrapped. 
     */
    private FileReader fileReader; 

    /* The root tags to add to elements or documents when helpful in
     * parsing.
     */ 
    private final String ROOT_OPEN = "<Root>", ROOT_CLOSE = "</Root>"; 


    public OracleReader(File oracleFile, Ignore ignore) throws IOException {
	super(); 
	this.oracleFile = oracleFile; 
	this.ignore = ignore; 
	initReader(); 
    }

    /* Opens the oracle-info file. Initializes stepNumber. Updates
     * stepStatus.
     */     
    public void initReader() throws IOException {
	stepStatus = STATUS_OK; 
	stepNumber = 0; 
	fileReader = new FileReader(oracleFile); 
	reader = new LineNumberReader(fileReader); 
    }

    /* Closes any open readers. 
     */
    protected void finalize() throws Throwable {
	if (reader != null) {
	    reader.close(); 
	}
	if (fileReader != null) {
	    fileReader.close(); 
	}
    }

    /* Reads and parses the next step in the oracle-info file if
     * stepStatus is STATUS_OK, skipping any material before the
     * '<Step>' tag. Updates stepStatus and stepNumber. 
     * 
     * A step is assumed to have the XML structure
     * '<Step>...</Step><Oracleinfo>...</Oracleinfo><Oracleinfo>...</Oracleinfo>'.
     * 
     * When searching for '<Step>' and '</Oracleinfo>', a naive text
     * search is used; the XML structure is ignored so that some steps
     * can be processed even if later in the oracle-info file a step
     * is incomplete. Each of these tags is assumed to appear on a
     * line by itself.
     */
    public Step nextStep() throws IOException, JDOMException, XMLStructureException {
	if (stepStatus == STATUS_OK) {
	    String line; 
	    do {
		line = reader.readLine(); 
		if (line == null) {
		    stepStatus = STATUS_ABSENT; 
		    return null; 
		}
	    } while (! line.contains("<Step>")); 
	    int stepLine = reader.getLineNumber(); 

	    /* It would be more elegant to implement this with piped
	     * streams than a temp file, but the limited buffer size
	     * of PipedInputStream makes this impractical.
	     */
	    File tmpFile = File.createTempFile("orc", null); 
	    FileWriter fw = new FileWriter(tmpFile); 
	    BufferedWriter bw = new BufferedWriter(fw); 
	    bw.write(ROOT_OPEN, 0, ROOT_OPEN.length());  bw.newLine();  // Must add a root so the document is well-formed
	    bw.write(line, 0, line.length());  bw.newLine(); 
	    int oracleInfos = 0; 
	    while (oracleInfos < 2) {
		line = reader.readLine(); 
		if (line == null) {
		    stepStatus = STATUS_PARTIAL; 
		    return null; 
		} else {
		    bw.write(line, 0, line.length());  bw.newLine(); 
		}
		if (line.contains("</Oracleinfo>")) {
		    oracleInfos++; 
		}
	    } 
	    bw.write(ROOT_CLOSE, 0, ROOT_CLOSE.length());  bw.newLine(); 
	    bw.close(); 
	    fw.close(); 
	    stepNumber++; 
	    SAXBuilder parser = new SAXBuilder(); 
	    Document doc = parser.build(tmpFile); 
	    tmpFile.delete(); 
	    return readStep(stepLine, doc.getRootElement()); 
	}
	return null; 
    }

    /* A step passed to this method is assumed to have the XML
     * structure
     * '<Step>...Event...</Step><Oracleinfo><GUI>...Component...</GUI>
     * </Oracleinfo><Oracleinfo>...Windows...</Oracleinfo>' enclosed in
     * tags ROOT_OPEN and ROOT_CLOSE. 
     * 
     * Properties in ignore are applied to both the component (first
     * <Oracleinfo>) and the windows (second <Oracleinfo>), but
     * classes and titles in ignore are applied only to the windows.
     */
    protected Step readStep(int stepLine, Element elem) throws JDOMException, XMLStructureException {
	if (elem.getChildren().size() != 3) {
	    throw new XMLStructureException("In step " + stepNumber + ", which starts at line " + stepLine); 
	}
	Element eventElem = (Element) elem.getChildren().get(0); 
	Element compElem = (Element) elem.getChildren().get(1); 
	Element winElem = (Element) elem.getChildren().get(2); 
	if (!eventElem.getName().equals("Step") || !compElem.getName().equals("Oracleinfo") || !winElem.getName().equals("Oracleinfo")) {
	    throw new XMLStructureException("In step " + stepNumber + ", which starts at line " + stepLine); 
	}

	Event event; ; 
	try {
	    event = readEvent(eventElem); 
	} catch (XMLStructureException e) {
	    throw new XMLStructureException("In step " + stepNumber + ", which starts at line " + stepLine, e); 
	}

	Component comp; 
	if (compElem.getChildren().size() != 1) {
	    throw new XMLStructureException("In step " + stepNumber + ", which starts at line " + stepLine); 
	}
	Element compChildElem = (Element) compElem.getChildren().get(0); 
	if (!compChildElem.getName().equals("GUI")) {
	    throw new XMLStructureException("In step " + stepNumber + ", which starts at line " + stepLine); 
	}
	if (compChildElem.getChildren().size() != 1) {
	    throw new XMLStructureException("In step " + stepNumber + ", which starts at line " + stepLine); 
	}
	Element compChildChildElem = (Element) compChildElem.getChildren().get(0); 
	try {
	    comp = readComponent(compChildChildElem, false); 
	} catch (XMLStructureException e) {
	    throw new XMLStructureException("In step " + stepNumber + ", which starts at line " + stepLine, e); 
	}

	Vector<Window> wins = new Vector<Window>(); 
	if (winElem.getChildren().size() == 0) {
	    throw new XMLStructureException("In step " + stepNumber + ", which starts at line " + stepLine); 
	}
	for (Object o : winElem.getChildren()) {
	    Element winChildElem = (Element) o; 
	    try {
		Window win = readWindow(winChildElem); 
		wins.addElement(win); 
	    } catch (XMLStructureException e) {
		throw new XMLStructureException("In step " + stepNumber + ", which starts at line " + stepLine, e); 
	    }
	}

	return new Step(stepNumber, stepLine, event, comp, wins); 
    }

    /* An event passed to this method is assumed to have the XML
     * structure
     * '<Step><Window>...</Window><Nonterminal>...</Nonterminal></Step>'.
     */
    protected Event readEvent(Element elem) throws JDOMException, XMLStructureException {
	if (!elem.getName().equals("Step")) {
	    throw new XMLStructureException(); 
	}
	Event event = new Event(); 
	for (Object o : elem.getChildren()) {
	    Element child = (Element) o; 
	    event.put(child.getName(), child.getText()); 
	}

	return event; 
    }

    /* A window passed to this method is assumed to have the XML
     * structure
     * '<GUI><Window><Attributes>...properties...</Attributes></Window>...components...</GUI>'.
     */
    protected Window readWindow(Element elem) throws JDOMException, XMLStructureException {
	if (!elem.getName().equals("GUI")) {
	    throw new XMLStructureException(); 
	}
	Window win = new Window(); 
	Element winChild = elem.getChild("Window"); 
	if (winChild.getChildren().size() != 1) {
	    throw new XMLStructureException(); 
	}
	Element winChildChild = (Element) winChild.getChildren().get(0); 
	if (!winChildChild.getName().equals("Attributes")) {
	    throw new XMLStructureException(); 
	}
	for (Object o : winChildChild.getChildren()) {
	    Element child = (Element) o; 
	    String[] nameValue = readProperty(child); 
	    if (!ignore.getProperties().contains(nameValue[0])) {
		win.put(nameValue[0], nameValue[1]); 
	    }
	}
	Vector<Component> comps = new Vector<Component>(); 
	for (Object o : elem.getChildren()) {
	    Element child = (Element) o; 
	    if (! child.getName().equals("Window")) {
		Component comp = readComponent(child, true); 
		if (comp != null) {
		    comps.addElement(comp); 
		}
	    }
	}
	win.setComponents(comps); 

	return win; 
    }

    /* A component passed to this method is assumed to have the XML
     * structure '<Component>...properties...</Component>'. Strips the
     * index suffix (e.g., '_R_10', '_N_1') from the component's
     * title. Returns null if possIgnore is set and the component's
     * (stripped) title or class is in ignore.
     */
    protected Component readComponent(Element elem, boolean possIgnore) throws JDOMException, XMLStructureException {
	if (!elem.getName().equals("Component")) {
	    throw new XMLStructureException(); 
	}
	Component comp = new Component(); 
	for (Object o : elem.getChildren()) {
	    Element child = (Element) o; 
	    String[] nameValue = readProperty(child); 
	    if (nameValue[0].equals("Title")) {
		int i = nameValue[1].lastIndexOf("_"); 
		i = nameValue[1].lastIndexOf("_", i - 1); 
		String suffix = nameValue[1].substring(i); 
		comp.setSuffix(suffix); 
		nameValue[1] = nameValue[1].substring(0, i); 
		if (possIgnore && ignore.getTitles().contains(nameValue[1])) {
		    return null; 
		}
	    } else if (nameValue[0].equals("Class")) {
		if (possIgnore && ignore.getClasses().contains(nameValue[1])) {
		    return null; 
		}
	    }
	    if (!ignore.getProperties().contains(nameValue[0])) {
		comp.put(nameValue[0], nameValue[1]); 
	    }
	}

	return comp; 
    }

    /* A property passed to this method is assumed to have the XML
     * structure
     * '<Property><Name>...</Name><Value>...</Value></Property>'.
     */
    protected String[] readProperty(Element elem) throws JDOMException, XMLStructureException {
	if (!elem.getName().equals("Property")) {
	    throw new XMLStructureException(); 
	}
	String[] nameValue = new String[2]; 
	List c = elem.getChildren(); 
	if (c.size() != 2) {
	    throw new XMLStructureException(); 
	}
	Element name = (Element) c.get(0); 
	Element value = (Element) c.get(1); 
	if (!name.getName().equals("Name") || !value.getName().equals("Value")) {
	    throw new XMLStructureException(); 
	}
	nameValue[0] = name.getText(); 
	nameValue[1] = value.getText(); 
	return nameValue; 
    }

    public int getStepStatus() {
	return stepStatus; 
    }

    public int getStepNumber() {
	return stepNumber; 
    }

    public class XMLStructureException extends Exception {

	public XMLStructureException() {
	    super(); 
	}

	public XMLStructureException(String s) {
	    super(s); 
	}

	public XMLStructureException(String s, Throwable t) {
	    super(s, t); 
	}

    }

    /* For testing. 
     */
    public static void main(String[] args) {
	try {
	    File oFile = new File(args[0]); 
	    File iFile = null; 
	    if (args.length > 1) {
		iFile = new File(args[1]); 
	    }
	    OracleReader o = new OracleReader(oFile, new Ignore(iFile)); 
	    Step s = o.nextStep(); 
	    while (s != null) {
		System.out.println(s.getNumber() + ":" + s.getLine() + " (" + o.getStepNumber() + ":" + o.getStepStatus() + ")"); 
		s = o.nextStep(); 
	    }
	} catch (Exception e) {
	    e.printStackTrace(); 
	}
    }

}
